from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min, CharField, Value as V
from django.db.models.functions import Concat
from NearBeach.forms import *
from NearBeach.views.tools.internal_functions import *
from NearBeach.user_permissions import return_user_permission_level


import json, urllib, urllib3

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_bug(request,destination,location_id):
    #ADD IN CHECK PERMISSIONS THAT USES THE DESTINATION AND LOCATION!

    # Get data from form
    form = AddBugForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    submit_bug = bug(
        bug_client=form.cleaned_data['bug_client'],
        bug_code=form.cleaned_data['bug_id'],
        bug_description=form.cleaned_data['bug_description'],
        bug_status=form.cleaned_data['bug_status'],
        change_user=request.user,
    )

    # Connect to the correct destination
    submit_bug = set_object_from_destination(submit_bug,destination,location_id)

    # Save
    submit_bug.save()

    # Get new bug to send back to use
    bug_results = bug.objects.filter(bug_id=submit_bug.bug_id)

    # Return the JSON data
    return HttpResponse(serializers.serialize('json',bug_results),content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_customer(request,destination,location_id):
    #ADD IN CHECK PERMISSIONS THAT USES THE DESTINATION AND LOCATION!

    # Get data from form
    form = AddCustomerForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Obtain the data dependent on the destination
    submit_object_assignment = object_assignment(
        change_user=request.user,
        customer=form.cleaned_data['customer']
    )
    submit_object_assignment = set_object_from_destination(
        submit_object_assignment,
        destination,
        location_id
    )

    # Save the data
    submit_object_assignment.save()

    customer_results = get_customer_list(destination, location_id)

    return HttpResponse(serializers.serialize('json', customer_results), content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_group(request,destination,location_id):
    # ADD IN CHECK PERMISSIONS THAT USES THE DESTINATION AND LOCATION!

    # Get data from form
    form = AddGroupForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # We loop through the responses and add them to the destination's object association
    group_list = request.POST.getlist('group_list')

    for single_group in group_list:
        # Get group instance
        group_instance = group.objects.get(group_id=single_group)

        # Construct the object assignment
        submit_object_assignment = object_assignment(
            group_id=group_instance,
            change_user=request.user,
        )
        submit_object_assignment = set_object_from_destination(submit_object_assignment,destination,location_id)

        # Save the data
        submit_object_assignment.save()

    # Get the new group list data
    group_results = get_group_list(destination,location_id)

    return HttpResponse(serializers.serialize('json',group_results),content_type='application/json')

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_link(request,destination,location_id):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    # ADD IN CHECKER FOR USER PERMISSIONS

    # Get the data
    form = AddObjectLinkForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Start saving the data
    object_assignment_submit = object_assignment(
        change_user=request.user,
    )

    # Add the destination/location_id to the object
    object_assignment_submit = link_object(object_assignment_submit,destination,location_id)

    # Loop through the results and add them in.
    for row in request.POST.getlist("project"):
        submit_object_assignment = object_assignment(
            project=project.objects.get(project_id=row),
            change_user=request.user,
        )
        set_object_from_destination(submit_object_assignment,destination,location_id)
        submit_object_assignment.save()

    for row in request.POST.getlist("task"):
        submit_object_assignment = object_assignment(
            task=task.objects.get(task_id=row),
            change_user=request.user,
        )
        set_object_from_destination(submit_object_assignment, destination, location_id)
        submit_object_assignment.save()

    for row in request.POST.getlist("requirement"):
        submit_object_assignment = object_assignment(
            requirement=requirement.objects.get(requirement_id=row),
            change_user=request.user,
        )
        set_object_from_destination(submit_object_assignment, destination, location_id)
        submit_object_assignment.save()

    for row in request.POST.getlist("requirement_item"):
        submit_object_assignment = object_assignment(
            requirement_item=requirement_item.objects.get(requirement_item_id=row),
            change_user=request.user,
        )
        set_object_from_destination(submit_object_assignment, destination, location_id)
        submit_object_assignment.save()


    return HttpResponse("Success")

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_notes(request,destination,location_id):
    # ADD IN PERMISSIONS HERE!

    # Fill out the form
    form = AddNoteForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # SAVE DATA
    submit_object_note = object_note(
        change_user = request.user,
        object_note = form.cleaned_data['note']
    )
    submit_object_note = set_object_from_destination(
        submit_object_note,
        destination,
        location_id
    )

    submit_object_note.save()

    # Get data to send back to user
    note_resuts = object_note.objects.filter(object_note_id=submit_object_note.object_note_id)

    return HttpResponse(serializers.serialize('json',note_resuts),content_type='application.json')

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def add_user(request,destination,location_id):
    # ADD IN A CHECK TO CHECK USER'S PERMISSION!

    # Check the data against the form
    form = AddUserForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Extract the list of users from the POST data
    user_list = request.POST.getlist('user_list')

    # Loop through them and add them to the object assignment
    for single_user in user_list:
        # Get user instance
        user_instance = User.objects.get(id=single_user)

        # Create object assignment
        submit_object_assignment = object_assignment(
            change_user=request.user,
            assigned_user=user_instance,
        )
        submit_object_assignment = set_object_from_destination(submit_object_assignment,destination,location_id)

        # Save
        submit_object_assignment.save()

    # Get the data to return to the user
    user_results = get_user_list(destination,location_id)

    return HttpResponse(serializers.serialize('json',user_results),content_type='application/json')

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def admin_add_user(request):
    """

    :param request:
    :return:
    """

    # Make sure user has permissions

    # Get data
    group_results = group.objects.filter(
        is_deleted=False,
    ).values()

    permission_set_results = permission_set.objects.filter(
        is_deleted=False,
    ).values()

    user_results = User.objects.filter(
        is_active=True,
    ).values()

    # Convert data to json format
    group_results = json.dumps(list(group_results), cls=DjangoJSONEncoder)
    permission_set_results = json.dumps(list(permission_set_results), cls=DjangoJSONEncoder)
    user_results = json.dumps(list(user_results), cls=DjangoJSONEncoder)

    return_data = {
        'group_results': json.loads(group_results),
        'permission_set_results': json.loads(permission_set_results),
        'user_results': json.loads(user_results),
    }

    return JsonResponse(return_data)


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def associated_objects(request,destination,location_id):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    # Organisations have a special method. We will return the results directly from this method to the user.
    if destination == 'organisation':
        return associated_objects_organisations(location_id)

    # Get the data
    object_assignment_results = object_assignment.objects.filter(
        is_deleted=False,
    )
    object_assignment_results = get_object_from_destination(object_assignment_results, destination, location_id)

    opportunity_results = opportunity.objects.filter(
        is_deleted=False,
        opportunity_id__in=object_assignment_results.filter(
            opportunity_id__isnull=False
        ).values('opportunity_id')
    ).values()

    project_results = project.objects.filter(
        is_deleted=False,
        project_id__in=object_assignment_results.filter(
            project_id__isnull=False
        ).values('project_id')
    ).values()

    requirement_results = requirement.objects.filter(
        is_deleted=False,
        requirement_id__in=object_assignment_results.filter(
            requirement_id__isnull=False
        ).values('requirement_id')
    ).values()

    task_results = task.objects.filter(
        is_deleted=False,
        task_id__in=object_assignment_results.filter(
            task_id__isnull=False
        ).values('task_id')
    ).values()

    # Return the JSON Response back - which will return strait to the user
    return JsonResponse({
        'opportunity': list(opportunity_results),
        'project': list(project_results),
        'requirement': list(requirement_results),
        'task': list(task_results),
    })


# Internal Functions
def associated_objects_organisations(location_id):
    """
    Due to organisation's links being connected to the objects directly. We will need to query all the objects that
    can be related to an organisation, and combine them into one JSON output.

    Please note - opportunities are the exception here - hence they are utilising the object assignment table.

    To make it JSON friendly, we have to add .values() to each object lookup, and then simple list them in the JSON
    return function below.
    :param location_id:
    :return:
    """
    # Get the data
    opportunity_results = opportunity.objects.filter(
        is_deleted=False,
        opportunity_id__in=object_assignment.objects.filter(
            is_deleted=False,
            opportunity_id__isnull=False,
            organisation_id=location_id,
        ).values('opportunity_id')
    ).values()

    project_results = project.objects.filter(
        is_deleted=False,
        organisation__in=location_id,
    ).values()

    requirement_results = requirement.objects.filter(
        is_deleted=False,
        organisation__in=location_id,
    ).values()

    task_results = task.objects.filter(
        is_deleted=False,
        organisation__in=location_id,
    ).values()

    # Return the JSON Response back - which will return strait to the user
    return JsonResponse({
        'opportunity': list(opportunity_results),
        'project': list(project_results),
        'requirement': list(requirement_results),
        'task': list(task_results),
    })


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def bug_client_list(request):
    bug_client_results = bug_client.objects.filter(
        is_deleted=False,
    )

    return HttpResponse(serializers.serialize('json',bug_client_results), content_type='application/json')

@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def bug_list(request,destination,location_id):
    # Obtain the data dependent on the destination
    bug_list = bug.objects.filter(
        is_deleted=False,
    )
    bug_list = get_object_from_destination(bug_list,destination,location_id)

    # Limit to certain values
    bug_list = bug_list.values(
        'bug_client',
        'bug_client__list_of_bug_client',
        'bug_client__list_of_bug_client__bug_client_name',
        'bug_client__bug_client_name',
        'bug_client__bug_client_url',
        'bug_code',
        'bug_description',
        'bug_status',
        "project_id",
        "requirement_id",
        "task_id",
    )

    """
    As explained on stack overflow here - https://stackoverflow.com/questions/7650448/django-serialize-queryset-values-into-json#31994176
    We need to Django's serializers can't handle a ValuesQuerySet. However, you can serialize by using a standard 
    json.dumps() and transforming your ValuesQuerySet to a list by using list().[sic]
    """

    # Send back json data
    json_results = json.dumps(list(bug_list), cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def customer_list(request,destination,location_id):
    customer_results = get_customer_list(destination, location_id)

    return HttpResponse(serializers.serialize('json', customer_results), content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def customer_list_all(request,destination,location_id):
    # Get the organisation dependant on the destination source
    if destination == "requirement":
        organisation_results = organisation.objects.get(
            organisation_id=requirement.objects.get(
                is_deleted=False,
                requirement_id=location_id,
            ).organisation_id
        )
    elif destination == "requirement_item":
        organisation_results = organisation.objects.get(
            organisation_id=requirement.objects.get(
                is_deleted=False,
                requirement_id=requirement_item.objects.get(
                    requirement_item_id=location_id
                ).requirement_id,
            ).organisation_id
        )
    elif destination == "project":
        organisation_results = organisation.objects.get(
            organisation_id=project.objects.get(
                is_deleted=False,
                project_id=location_id,
            ).organisation_id
        )
    elif destination == "task":
        organisation_results = organisation.objects.get(
            organisation_id=task.objects.get(
                is_deleted=False,
                task_id=location_id,
            ).organisation_id
        )
    else:
        # There is no destination that could match this. Send user to errors
        return HttpResponseBadRequest("Sorry - there was an error getting the Customer List")

    customer_results = customer.objects.filter(
        is_deleted=False,
        organisation_id=organisation_results.organisation_id
    )

    return HttpResponse(serializers.serialize('json',customer_results), content_type='application/json')

# Internal function
def get_customer_list(destination,location_id):
    # Get a list of all objects assignments dependant on the destination
    object_customers = object_assignment.objects.filter(
        is_deleted=False,
        customer_id__isnull=False,
    )
    object_customers = get_object_from_destination(object_customers,destination,location_id)

    return customer.objects.filter(
        is_deleted=False,
        customer_id__in=object_customers.values('customer_id')
    )

# Internal function
def get_group_list(destination,location_id):
    object_results = object_assignment.objects.filter(
        is_deleted=False,
    )
    object_results = get_object_from_destination(
        object_results,
        destination,
        location_id
    )

    # Now return the groups
    return group.objects.filter(
        is_deleted=False,
        group_id__in=object_results.values('group_id')
    )


# Internal Function
def get_user_list(destination,location_id):
    # Get the data we want
    object_results = object_assignment.objects.filter(
        is_deleted=False,
        assigned_user_id__isnull=False,
    )
    object_results = get_object_from_destination(
        object_results,
        destination,
        location_id
    )

    # Get the user details
    return User.objects.filter(id__in=object_results.values('assigned_user_id'))


# Internal Function
def get_user_list_all(destination,location_id):
    # Get a list of users we want to exclude
    object_results = object_assignment.objects.filter(
        is_deleted=False,
        assigned_user_id__isnull=False,
    )
    object_results = get_object_from_destination(
        object_results,
        destination,
        location_id
    )

    # Get a list of all the groups associated with this destination
    group_results = object_assignment.objects.filter(
        is_deleted=False,
        group_id__isnull=False,
    )
    group_results = get_object_from_destination(group_results,destination,location_id)

    # Get a list of users who are associated with these groups & not in the excluded list
    user_results = User.objects.filter(
        id__in=user_group.objects.filter(
            is_deleted=False,
            group_id__in=group_results.values('group_id'),
        ).values('username_id'),
        is_active=True,
    ).exclude(
        id__in=object_results.values('assigned_user_id')
    )

    # Send the results back
    return user_results


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def group_list(request,destination,location_id):
    # Get the data dependant on the object lookup
    group_results = get_group_list(destination,location_id)

    # Return the data
    return HttpResponse(serializers.serialize('json',group_results),content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def group_list_all(request,destination,location_id):
    # ADD CHECKS FOR USER PERMISSIONS!

    # Obtain data
    group_existing_results = object_assignment.objects.filter(
        is_deleted=False,
        group_id__isnull=False,
    )
    group_existing_results = get_object_from_destination(group_existing_results,destination,location_id)

    group_results = group.objects.filter(
        is_deleted=False,
    ).exclude(
        group_id__in=group_existing_results.values('group_id')
    )

    # Return data as json
    return HttpResponse(serializers.serialize('json',group_results),content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def lead_user_list(request):
    """

    :param request:
    :return:
    """

    # Get the data
    search_form = SearchForm(request.POST)
    if not search_form.is_valid():
        return HttpResponseBadRequest(search_form.errors)

    # First we create a search string and annotate it onto our results
    user_results = User.objects.annotate(
        search_string=Concat(
            'username', V(' '), 'first_name', V(' '), 'last_name', V(' '), 'email',
            output_field=CharField()
        )
    ).filter(
        is_active=True,
    )

    for split_row in search_form.cleaned_data['search'].split(' '):
        """
        """
        user_results.filter(
            search_string__icontains=split_row,
        )

    # Return the json data
    return HttpResponse(serializers.serialize('json', user_results[:25]), content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def link_list(request,destination,location_id,object_lookup):
    # Get the data dependent on the object lookup
    if object_lookup == 'project':
        data_results = project.objects.filter(
            is_deleted=False,
        ).exclude(
            project_status='Closed',
        )
    elif object_lookup == "task":
        data_results = task.objects.filter(
            is_deleted=False,
        ).exclude(
            task_status='Closed',
        )
    elif object_lookup == "requirement":
        data_results = requirement.objects.filter(
            is_deleted=False,
            requirement_status_id__in=list_of_requirement_status.objects.filter(
                is_deleted=False,
                requirement_status_is_closed=False,
            ).values('requirement_status_id'),
        )
    elif object_lookup == "requirement_item":
        data_results = requirement_item.objects.filter(
            is_deleted=False,
            requirement_item_status_id__in=list_of_requirement_item_status.objects.filter(
                is_deleted=False,
                status_is_closed=False,
            ).values('requirement_item_status_id'),
            requirement_id__in=requirement.objects.filter(
                is_deleted=False,
                requirement_status_id__in=list_of_requirement_status.objects.filter(
                    is_deleted=False,
                    requirement_status_is_closed=False,
                ).values('requirement_status_id')
            ).values('requirement_id'),
        )
    else:
        # There is an error.
        return HttpResponseBadRequest("Sorry - but that object lookup does not exist")

    # Send the data to the user
    return HttpResponse(serializers.serialize('json',data_results), content_type='application/json')


# Internal function
def link_object(object_assignment_submit,destination,location_id):
    """
    This is an internal function - depending on the destination, depends on what we are linking in the
    object_association_submit
    :param object_assignment_submit:
    :param destination:
    :param location_id:
    :return:
    """
    if destination == "project":
        object_assignment_submit.project = project.objects.get(project_id=location_id)
    elif destination == "requirement":
        object_assignment_submit.requirement = requirement.objects.get(requirement_id=location_id)
    elif destination == "requirement_item":
        object_assignment_submit.requirement_item = requirement_item.objects.get(requirement_item_id=location_id)
    elif destination == "task":
        object_assignment_submit.task = task.objects.get(task_id=location_id)

    # Return the results
    return object_assignment_submit


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def note_list(request,destination,location_id):
    # Everyone should have access to the notes section.

    # Get the notes dependent on the user destination and location
    note_results = object_note.objects.filter(
        is_deleted=False,
    )

    # Filter by destination and location_id
    note_results = get_object_from_destination(note_results,destination,location_id)

    # Return JSON results
    return HttpResponse(serializers.serialize('json',note_results),content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def object_link_list(request,destination,location_id):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """
    object_assignment_results = object_assignment.objects.filter(
        is_deleted=False,
    )
    object_assignment_results = get_object_from_destination(object_assignment_results,destination,location_id)

    object_assignment_results = object_assignment_results.filter(
        Q(project__isnull=False) |
        Q(requirement__isnull=False) |
        Q(requirement_item__isnull=False) |
        Q(task__isnull=False)
    ).values(
        'project_id',
        'project_id__project_name',
        'project_id__project_status',
        'task_id',
        'task_id__task_short_description',
        'task_id__task_status',
        'requirement_id',
        'requirement_id__requirement_title',
        'requirement_id__requirement_status__requirement_status',
        'requirement_item_id',
        'requirement_item_id__requirement_item_title',
        'requirement_item_id__requirement_item_status__requirement_item_status',
    )

    """
    As explained on stack overflow here - https://stackoverflow.com/questions/7650448/django-serialize-queryset-values-into-json#31994176
    We need to Django's serializers can't handle a ValuesQuerySet. However, you can serialize by using a standard 
    json.dumps() and transforming your ValuesQuerySet to a list by using list().[sic]
    """

    # Send back json data
    json_results = json.dumps(list(object_assignment_results), cls=DjangoJSONEncoder)

    return HttpResponse(json_results, content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def query_bug_client(request,destination,location_id):
    # Insert data into form
    form = QueryBugClientForm(request.POST)

    # Check to make sure everything is fine with the form
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Extract the information from the form
    bug_client_instance = form.cleaned_data['bug_client_id']
    search_terms = form.cleaned_data['search']

    # Get existing bugs that we want to extract out
    existing_bugs = bug.objects.filter(
        is_deleted=False,
        bug_client_id=bug_client_instance.bug_client_id,
    )
    existing_bugs = get_object_from_destination(existing_bugs,destination,location_id)

    # The values in the URL
    f_bugs = ''
    o_notequals = ''
    v_values = ''

    # The for loop
    for idx, row in enumerate(existing_bugs):
        nidx = str(idx + 1)
        f_bugs = f_bugs + "&f" + nidx + "=bug_id"
        o_notequals = o_notequals + "&o" + nidx + "=notequals"
        v_values = v_values + "&v" + nidx + "=" + str(row.bug_code)

    exclude_url = f_bugs + o_notequals + v_values

    url = bug_client_instance.bug_client_url \
          + bug_client_instance.list_of_bug_client.bug_client_api_url \
          + bug_client_instance.list_of_bug_client.api_search_bugs \
          + urllib.parse.quote(form.cleaned_data['search']) \
          + exclude_url

    """
    SECURITY ISSUE
    ~~~~~~~~~~~~~~
    The URL could contain a file. Which we do not want executed by mistake. So we just make sure that the URL starts
    with a http instead of ftp or file.

    We place the  at the end of the json_data because we have checked the field. This should be just a json 
    response. If it is not at this point then it will produce a server issue.
    """
    if url.lower().startswith('http'):
        # setup the pool manager for urllib3
        http = urllib3.PoolManager()

        # Plug in the url
        r = http.request('GET',url)

        # Extract the data
        json_data = json.loads(r.data.decode('utf-8'))
    else:
        raise ValueError from None

    # Send back the JSON data
    return JsonResponse(json_data['bugs'], safe=False)


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def user_list(request,destination,location_id):
    # Get the data we want
    user_results = get_user_list(destination,location_id)

    return HttpResponse(serializers.serialize('json',user_results),content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def user_list_all(request,destination,location_id):
    # ADD IN PERMISSIONS LATER

    # Get Data we want
    user_results = get_user_list_all(destination,location_id)

    return HttpResponse(serializers.serialize('json',user_results),content_type='application/json')



# Import the locust libraries
import time, getpass
from locust import HttpUser, task, between

#Request the username and password to login into NearBeach
username = input("\nPlease enter the username: ")
password = getpass.getpass("\nPlease enter the password: ")

class QuickstartUser(HttpUser):
    wait_time = between(1, 2)

    def on_start(self):
        """ on_start is called when a Locust start before any task is scheduled """
        self.client.post("/login", json={"id_username": username, "id_password": password})

    @task
    def dashboard(self):
        self.client.get("/")
       

    @task(3)
    def requirement_information(self):
        for item_id in range(10):
            self.client.get(f"/requirement_information/1")






# Generated by Django 3.1.2 on 2020-10-24 01:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NearBeach', '0002_initialise_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='about_user',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='document',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='group_permission',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='kanban_board',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='kanban_card',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='kanban_column',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='kanban_comment',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='kanban_level',
            name='is_deleted',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.1.2 on 2021-02-27 06:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('NearBeach', '0005_auto_20210123_1537'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='quote_product_and_service',
            name='change_user',
        ),
        migrations.RemoveField(
            model_name='quote_product_and_service',
            name='product_and_service',
        ),
        migrations.RemoveField(
            model_name='quote_product_and_service',
            name='quote',
        ),
        migrations.RemoveField(
            model_name='quote_product_and_service',
            name='tax',
        ),
        migrations.RemoveField(
            model_name='quote_responsible_customer',
            name='change_user',
        ),
        migrations.RemoveField(
            model_name='quote_responsible_customer',
            name='customer',
        ),
        migrations.RemoveField(
            model_name='quote_responsible_customer',
            name='quote',
        ),
        migrations.RemoveField(
            model_name='quote_template',
            name='change_user',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='assign_campus_to_customer',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='associate_project_and_task',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='bug',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='contact_history',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='email',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='invoice',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='invoice_product',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='opportunity',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='organisation_campus',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='quote',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='requirement_link',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='tag',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='tax',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='template',
        ),
        migrations.RemoveField(
            model_name='permission_set',
            name='whiteboard',
        ),
        migrations.DeleteModel(
            name='product_and_service',
        ),
        migrations.DeleteModel(
            name='quote_product_and_service',
        ),
        migrations.DeleteModel(
            name='quote_responsible_customer',
        ),
        migrations.DeleteModel(
            name='quote_template',
        ),
    ]



"""
Django Settings for TESTING PURPOSES

Do not utilise this settings.py file for your own project. Even if it is not
a production environment.

This file is only for the automatic testing and is not build for server use.
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'jz0k8%ecl#k!z+(9+5(^do1w!11ysus21m41m@i9c#u)*vk($o'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'NearBeach.apps.NearBeachConfig',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'NearBeach.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')]
        ,
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


#WSGI_APPLICATION = 'untitled.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.mysql',
#         'NAME': 'nearbeach_test',
#         'USER': 'root',
#         'PASSWORD': '',
#         'HOST': 'localhost',
#         'PORT': '3306',
#     }
# }


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


PROJECT_PATH = os.path.abspath(os.path.dirname(__file__))
PRIVATE_MEDIA_ROOT = os.path.join(PROJECT_PATH, 'private')
PRIVATE_MEDIA_SERVER = 'DefaultServer'
PRIVATE_MEDIA_URL = '/private/'

STATIC_URL = '/static/'
STATIC_ROOT= os.path.join(BASE_DIR,'static/')

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR,'media/')



from django.http import HttpResponseRedirect
from django.core.exceptions import PermissionDenied
from django.contrib.auth.decorators import user_passes_test
from django.db.models import Q, Max
from functools import wraps

from NearBeach.models import *


def project_permissions(min_permission_level):
    def decorator(func):
        @wraps(func)
        def inner(request, *args, **kwargs):
            # If user is admin - grant them all permissions
            if request.user.is_superuser:
                return func(request, *args, **kwargs)

            # Default user level is 0
            user_group_results = user_group.objects.filter(
                is_deleted=False,
                username=request.user,
            )

            # If we are passing the project_id through, we will use a different function
            if len(kwargs) > 0:
                # Check user permissions AGAINST the project and it's groups
                project_instance = project.objects.get(project_id=kwargs['project_id'])

                # Determine if there are any cross over with user groups and project groups
                group_results = group.objects.filter(
                    Q(
                        is_deleted=False,
                        # The project groups
                        group_id__in=object_assignment.objects.filter(
                            is_deleted=False,
                            project_id=kwargs['project_id'],
                        ).values('group_id'),
                    ) &
                    Q(
                        group_id__in=user_group_results.values('group_id')
                    )
                )

                # Check to make sure the user groups intersect
                if len(group_results) == 0:
                    # There are no matching groups - i.e. the user does not have any permission
                    raise PermissionDenied

            # Get the max permission value from user_group_results
            user_level = user_group_results.aggregate(
                Max('permission_set__project')
            )['permission_set__project__max']

            if user_level >= min_permission_level:
                # Everything is fine - continue on
                return func(request, *args, **kwargs)

            # Does not meet conditions
            raise PermissionDenied
        return inner
    return decorator



from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect
from django.urls import reverse
from django.template import loader
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods
from django.core import serializers

from NearBeach.models import *
from NearBeach.forms import NewRequestForChangeForm, RfcModuleForm, RfcInformationSaveForm, NewChangeTaskForm, \
    UpdateRFCStatus


# Internal function
def get_rfc_context(rfc_id):
    """
    Using the rfc ID, it will extract the required context for the rfc_information and rfc_readonly
    :param rfc_id:
    :return:
    """
    # Get data
    rfc_results = request_for_change.objects.get(rfc_id=rfc_id)
    rfc_change_lead = User.objects.get(id=rfc_results.rfc_lead.id)
    user_list = User.objects.filter(
        is_active=True,
        id__in=user_group.objects.filter(
            is_deleted=False,
            group_id__in=object_assignment.objects.filter(
                is_deleted=False,
                request_for_change_id=rfc_id,
            ).values('group_id')
        ).values('username_id')
    )

    # Context
    c = {
        'rfc_id': rfc_id,
        'rfc_results': serializers.serialize('json', [rfc_results]),
        'rfc_change_lead': serializers.serialize('json', [rfc_change_lead]),
        'user_list': serializers.serialize('json', user_list),
    }

    return c


@login_required(login_url='login', redirect_field_name="")
def new_request_for_change(request):
    """

    :param request:
    :return:
    """

    # CHECK USER PERMISSIONS

    # Get template
    t = loader.get_template('NearBeach/request_for_change/new_request_for_change.html')

    # Get data
    group_results = group.objects.filter(
        is_deleted=False,
    )

    user_results = User.objects.filter(  # This should only be group leaders
        is_active=True,
    )

    # Context
    c = {
        'group_results': serializers.serialize('json', group_results),
        'user_results': serializers.serialize('json', user_results),
    }

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def new_request_for_change_save(request):
    """

    :param request:
    :return:
    """

    # Check the user's permission

    # Get the form data
    form = NewRequestForChangeForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    rfc_submit = request_for_change(
        change_user=request.user,
        creation_user=request.user,
        rfc_status=1,
        rfc_title=form.cleaned_data['rfc_title'],
        rfc_summary=form.cleaned_data['rfc_summary'],
        rfc_type=form.cleaned_data['rfc_type'],
        rfc_implementation_start_date=form.cleaned_data['rfc_implementation_start_date'],
        rfc_implementation_end_date=form.cleaned_data['rfc_implementation_end_date'],
        rfc_implementation_release_date=form.cleaned_data['rfc_implementation_release_date'],
        rfc_version_number=form.cleaned_data['rfc_version_number'],
        rfc_lead=form.cleaned_data['rfc_lead'],
        rfc_priority=form.cleaned_data['rfc_priority'],
        rfc_risk=form.cleaned_data['rfc_risk'],
        rfc_impact=form.cleaned_data['rfc_impact'],
        rfc_risk_and_impact_analysis=form.cleaned_data['rfc_risk_and_impact_analysis'],
        rfc_implementation_plan=form.cleaned_data['rfc_implementation_plan'],
        rfc_backout_plan=form.cleaned_data['rfc_backout_plan'],
        rfc_test_plan=form.cleaned_data['rfc_test_plan'],
    )
    rfc_submit.save()

    # Get the group list and apply the permissions
    group_list = request.POST.getlist("group_list")

    for single_group in group_list:
        # Get the group instance
        group_instance = group.objects.get(group_id=single_group)

        # Save the group instance against object assignment
        submit_object_assignment = object_assignment(
            group_id=group_instance,
            request_for_change=rfc_submit,
            change_user=request.user,
        )

        # Save
        submit_object_assignment.save()

    # Send back requirement_information URL
    return HttpResponse(reverse('rfc_information', args={rfc_submit.rfc_id}))


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_change_task_list(request, rfc_id):
    """
    """
    change_task_results = change_task.objects.filter(
        is_deleted=False,
        request_for_change=rfc_id,
    ).order_by('change_task_start_date', 'change_task_end_date')

    # Send back JSON response
    return HttpResponse(serializers.serialize('json', change_task_results), content_type='application/json')


@login_required(login_url='login', redirect_field_name="")
def rfc_deployment(request, rfc_id):
    """

    :param request:
    :param rfc_id:
    :return:
    """
    # If rfc is not in draft mode - send user away
    rfc_results = request_for_change.objects.get(rfc_id=rfc_id)
    if not rfc_results.rfc_status == 3:  # Approved
        return HttpResponseRedirect(reverse('rfc_readonly', args={rfc_id}))

    # Get template
    t = loader.get_template('NearBeach/request_for_change/rfc_deployment.html')

    # Get context
    c = get_rfc_context(rfc_id)

    return HttpResponse(t.render(c, request))



@require_http_methods(["POST"])
@login_required(login_url='login', redirect_field_name="")
def rfc_new_change_task(request, rfc_id):
    """

    :param request:
    :param rfc_id:
    :return:
    """

    # ADD IN USER PERMISSIONS

    # Place data into forms for validation
    form = NewChangeTaskForm(request.POST)
    if not form.is_valid():
        # Send user bad request
        return HttpResponseBadRequest(form.errors)

    # Save the data
    submit_change_task = change_task(
        request_for_change=form.cleaned_data['request_for_change'],
        change_task_title=form.cleaned_data['change_task_title'],
        change_task_description=form.cleaned_data['change_task_description'],
        change_task_start_date=form.cleaned_data['change_task_start_date'],
        change_task_end_date=form.cleaned_data['change_task_end_date'],
        change_task_seconds=form.cleaned_data['change_task_seconds'],
        # change_task_assigned_user = form.cleaned_data['change_task_assigned_user'],
        # change_task_qa_user = form.cleaned_data['change_task_qa_user'],
        change_task_assigned_user=request.user,
        change_task_qa_user=request.user,
        change_task_required_by=form.cleaned_data['change_task_required_by'],
        is_downtime=form.cleaned_data['is_downtime'],
        change_task_status=1,
        change_user=request.user,
        creation_user=request.user,
    )
    submit_change_task.save()

    # Send back all the RFC change items
    change_item_results = change_task.objects.filter(
        is_deleted=False,
        request_for_change_id=rfc_id,
    )

    return HttpResponse(serializers.serialize('json', change_item_results), content_type='application/json')


@login_required(login_url='login', redirect_field_name="")
def rfc_information(request, rfc_id):
    """

    :param request:
    :param rfc_id:
    :return:
    """

    # If rfc is not in draft mode - send user away
    rfc_results = request_for_change.objects.get(rfc_id=rfc_id)
    if not rfc_results.rfc_status == 1:  # Draft
        return HttpResponseRedirect(reverse('rfc_readonly', args={rfc_id}))

    # Get template
    t = loader.get_template('NearBeach/request_for_change/rfc_information.html')

    # Get context
    c = get_rfc_context(rfc_id)

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_information_save(request, rfc_id):
    """

    :param request:
    :param rfc_id:
    :return:
    """

    # PROGRAM IN PERMISSIONS

    # Get the form data
    form = RfcInformationSaveForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the request for change data
    update_rfc = request_for_change.objects.get(rfc_id=rfc_id)

    # Update the data
    update_rfc.rfc_title = form.cleaned_data['rfc_title']
    update_rfc.rfc_summary = form.cleaned_data['rfc_summary']
    update_rfc.rfc_type = form.cleaned_data['rfc_type']
    update_rfc.rfc_version_number = form.cleaned_data['rfc_version_number']
    update_rfc.rfc_implementation_start_date = form.cleaned_data['rfc_implementation_start_date']
    update_rfc.rfc_implementation_end_date = form.cleaned_data['rfc_implementation_end_date']
    update_rfc.rfc_implementation_release_date = form.cleaned_data['rfc_implementation_release_date']

    # Save the data
    update_rfc.save()

    # Return blank success
    return HttpResponse("")


@login_required(login_url='login', redirect_field_name="")
def rfc_readonly(request, rfc_id):
    """

    :param request:
    :param rfc_id:
    :return:
    """

    # Get template
    t = loader.get_template('NearBeach/request_for_change/rfc_readonly.html')

    # Get context
    c = get_rfc_context(rfc_id)

    # Determine if the user is a group leader for ANY of the groups assigned to this rfc
    group_leader_count = object_assignment.objects.filter(
        is_deleted=False,
        request_for_change_id=rfc_id,
        group_id__in=user_group.objects.filter(
            is_deleted=False,
            username_id=request.user,
            group_leader=True,
        ).values('group_id')
    ).count()

    # Add the group_leader_count to c dict
    c.update({'group_leader_count': group_leader_count})

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_save_backout(request, rfc_id):
    """

    :param request:
    :return:
    """

    # Get the form data
    form = RfcModuleForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the RFC in question
    update_rfc = request_for_change.objects.get(rfc_id=rfc_id)

    # Update the rfc
    update_rfc.rfc_backout_plan = form.cleaned_data['text_input']

    # Save
    update_rfc.save()

    return HttpResponse("")


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_save_implementation(request, rfc_id):
    """
    """

    # Check user permissions

    # Get the form data
    form = RfcModuleForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the rfc in question
    update_rfc = request_for_change.objects.get(rfc_id=rfc_id)

    # Update the data
    update_rfc.rfc_implementation_plan = form.cleaned_data['text_input']

    # Save
    update_rfc.save()

    return HttpResponse("")


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name='')
def rfc_save_risk(request, rfc_id):
    """
    """

    # CHECK USER PERMISSIONS

    # Get the form data
    form = RfcModuleForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the RFC in question
    update_rfc = request_for_change.objects.get(rfc_id=rfc_id)

    # Fill in the data
    update_rfc.rfc_priority = form.cleaned_data['priority_of_change']
    update_rfc.rfc_risk = form.cleaned_data['risk_of_change']
    update_rfc.rfc_impact = form.cleaned_data['impact_of_change']
    update_rfc.rfc_risk_and_impact_analysis = form.cleaned_data['text_input']

    # Save the data
    update_rfc.save()

    # Return blank result
    return HttpResponse("")


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_save_test(request, rfc_id):
    """
    """

    # Check user permissions

    # Get the form data
    form = RfcModuleForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the rfc in question
    update_rfc = request_for_change.objects.get(rfc_id=rfc_id)

    # Update the rfc's data
    update_rfc.rfc_test_plan = form.cleaned_data['text_input']

    # Save data
    update_rfc.save()

    return HttpResponse("")


# Internal function
def rfc_status_approved(rfc_id, rfc_results, request):
    """
    Method
    ~~~~~~
    1. Gather all User's rfc_group_rfc_approvals - and approve
    2. Check if there are any waiting rfc_group_rfc_approvals left - if none, approve rfc
    :param request:
    :param rfc_results:
    :param rfc_id:
    :return:
    """
    # Get group results
    group_results = user_group.objects.filter(
        is_deleted=False,
        username=request.user,
    )

    # Update all user's rfc_group_approvals to approved
    request_for_change_group_approval.objects.filter(
        is_deleted=False,
        rfc_id=rfc_id,
        group_id__in=group_results.values('group_id')
    ).update(approval=2)

    # Send off to check to make sure that the rfc status needs updating
    rfc_status_check_approval_status(rfc_id, rfc_results, group_results)

    return


# Internal function
def rfc_status_check_approval_status(rfc_id, rfc_results, group_results):
    """

    :param group_results:
    :param rfc_results:
    :param rfc_id:
    :return:
    """

    # Check all submitted group approvals to make sure that they are all approved - if they are, update the status.
    non_approved_group_approvals = request_for_change_group_approval.objects.filter(
        is_deleted=False,
        group_id__in=group_results.values('group_id'),
        rfc_id=rfc_id,
        approval=1,
    ).count()

    # If there are no waiting for approval results - we default up to approved
    if non_approved_group_approvals == 0:
        rfc_results.rfc_status = 3  # Approved value
        rfc_results.save()

        # Update all change tasks to approved
        change_task.objects.filter(
            is_deleted=False,
            request_for_change_id=rfc_id,
            change_task_status=2,
        ).update(change_task_status=3)

    return


# Internal function
def rfc_status_rejected(rfc_id, rfc_results):
    """

    :param rfc_results:
    :param rfc_id:
    :return:
    """
    # Update all user's rfc_group_approvals to approved
    request_for_change_group_approval.objects.filter(
        is_deleted=False,
        rfc_id=rfc_id,
    ).update(approval=3)

    # Reject who RFC
    rfc_results.rfc_status = 6
    rfc_results.save()

    # Reject all the change tasks
    change_task.objects.filter(
        is_deleted=False,
        rfc_id=rfc_id,
    ).update(change_task_status=6)

    return


# Internal function
def rfc_status_waiting_for_approval(rfc_id, rfc_results, request):
    """
    Method
    ~~~~~~
    1. Find all groups associated with the RFC
    2. Create a single row for all groups in the request_for_change_group_approval
    3. Check to see if there are any group leaders in said groups - or auto approve
    4. Check to see if all tasks are auto approved, if so - auto apprve the rfc :)
    :param request:
    :param rfc_results:
    :param rfc_id:
    :return:
    """

    # Get the group results
    group_results = group.objects.filter(
        is_deleted=False,
        group_id__in=object_assignment.objects.filter(
            is_deleted=False,
            group_id__isnull=False,
            request_for_change_id=rfc_id,
        ).values('group_id')
    )

    # Place all change tasks into waiting
    change_task.objects.filter(
        is_deleted=False,
        request_for_change_id=rfc_id,
        change_task_status=1,
    ).update(change_task_status=2)

    # Loop through the groups, create the group approval, and see if there are ANY group leaders
    for single_group in group_results:
        # Create the group_approval
        submit_group_approval = request_for_change_group_approval(
            rfc_id=rfc_id,
            group_id=single_group.group_id,
            change_user_id=request.user.id,
        )

        # Check to see if there are any group leaders
        group_leader_count = user_group.objects.filter(
            is_deleted=False,
            group_id=single_group.group_id,
            group_leader=True,
        ).count()

        if group_leader_count == 0:
            submit_group_approval.approval = 2
        else:
            submit_group_approval.approval = 1

        # Save the data
        submit_group_approval.save()

    rfc_status_check_approval_status(rfc_id, rfc_results, group_results)

    return


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def rfc_update_status(request, rfc_id):
    """
    Using a simple form, we determine which status we are going to update to - and apply the correct status.
    :param request:
    :param rfc_id:
    :return:
    """

    # Add in user permissions

    # Get the form data
    form = UpdateRFCStatus(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the rfc
    rfc_update = request_for_change.objects.get(rfc_id=rfc_id)

    # Update the status
    rfc_update.rfc_status = int(form.cleaned_data['rfc_status'])
    rfc_update.save()

    # Depending on what the status is depends what to do
    status_dict = dict(RFC_STATUS)
    if status_dict[form.cleaned_data['rfc_status']] == 'Waiting for approval':
        rfc_status_waiting_for_approval(rfc_id, rfc_update, request)
    if status_dict[form.cleaned_data['rfc_status']] == 'Approved':
        rfc_status_approved(rfc_id, rfc_update, request)
    if status_dict[form.cleaned_data['rfc_status']] == 'Rejected':
        rfc_status_rejected(rfc_id, rfc_update)

    return HttpResponse("")



from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.template import loader
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse, HttpResponseNotFound
from django.views.decorators.http import require_http_methods
from django.db.models import Q

# Import Forms
from NearBeach.forms import *

import json


# Internal Function
def get_object_search_data(search_form):
    """
    The following internal function will search the following objects using the form's data;
    - Request for Change
    - Requirements
    - Projects
    - Tasks
    It will combine them into a single JSON array and send back to the previous function
    :param form: Contains all the data we require.
    :return:
    """

    # Get instance data for all objects
    rfc_results = request_for_change.objects.filter(is_deleted=False).values(
        'rfc_id',
        'rfc_title',
        'rfc_status',
    )
    requirement_results = requirement.objects.filter(is_deleted=False).values(
        'requirement_id',
        'requirement_title',
        'requirement_status__requirement_status',
    )
    project_results = project.objects.filter(is_deleted=False).values(
        'project_id',
        'project_name',
        'project_status',
    )
    task_results = task.objects.filter(is_deleted=False).values(
        'task_id',
        'task_short_description',
        'task_status',
    )
    kanban_results = kanban_board.objects.filter(is_deleted=False).values(
        'kanban_board_id',
        'kanban_board_name',
        'kanban_board_status',
    )

    # Check to see if we are searching for closed objects
    include_closed = search_form.cleaned_data['include_closed']

    # If we are NOT including closed - then we will limit to those with status is_deleted=False
    if not include_closed:
        ### NEED TO PROGRAM IN THE REAL CLOSED VALUES!!! ###
        rfc_results = rfc_results.filter(is_deleted=False)
        requirement_results = requirement_results.filter(is_deleted=False)
        project_results = project_results.filter(is_deleted=False)
        task_results = task_results.filter(is_deleted=False)
        kanban_results = kanban_results.filter(is_deleted=False)

    # Split the space results - then apply the filter of each split value
    for split_row in search_form.cleaned_data['search'].split(' '):
        # Update the each instance with the split row results
        rfc_results = rfc_results.filter(
            Q(rfc_title__icontains=split_row)
        )
        requirement_results = requirement_results.filter(
            Q(requirement_title__icontains=split_row)
        )
        project_results = project_results.filter(
            Q(project_name__icontains=split_row)
        )
        task_results = task_results.filter(
            Q(task_short_description__icontains=split_row)
        )
        kanban_results = kanban_results.filter(
            Q(kanban_board_name__icontains=split_row)
        )

        # If the split row is a number - also check against the id
        if split_row.isnumeric():
            rfc_results = rfc_results.filter(
                Q(rfc_id=split_row)
            )
            requirement_results = requirement_results.filter(
                Q(requirement_id=split_row)
            )
            project_results = project_results.filter(
                Q(project_id=split_row)
            )
            task_results = task_results.filter(
                Q(task_id=split_row)
            )
            kanban_results = kanban_results.filter(
                Q(kanban_board_id=split_row)
            )


    # Only have 25 results and order by alphabetical order
    rfc_results.order_by('rfc_title')[:25]
    requirement_results.order_by('requirement_title')[:25]
    project_results.order_by('project_name')[:25]
    task_results.order_by('task_short_description').values()[:25]
    kanban_results.order_by('kanban_board_name').values()[:25]

    """
    The pain point
    ~~~~~~~~~~~~~~
    Due to Django wanting to send converted json data as a string, we have to;
    1. Apply serialisation
    2. Apply a json.loads function
    3. Compile data and send back.
    
    Note to Django developers - there has to be a better way
    """
    rfc_results = json.dumps(list(rfc_results), cls=DjangoJSONEncoder)
    requirement_results = json.dumps(list(requirement_results), cls=DjangoJSONEncoder)
    project_results = json.dumps(list(project_results), cls=DjangoJSONEncoder)
    task_results = json.dumps(list(task_results), cls=DjangoJSONEncoder)
    kanban_results = json.dumps(list(kanban_results), cls=DjangoJSONEncoder)

    # Send back a JSON array with JSON arrays inside
    return {
        'request_for_change': json.loads(rfc_results),
        'requirement': json.loads(requirement_results),
        'project': json.loads(project_results),
        'task': json.loads(task_results),
        'kanban': json.loads(kanban_results),
    }


@login_required(login_url='login',redirect_field_name="")
def search(request):
    """

    :param request:
    :return:
    """
    form = SearchObjectsForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Template
    t = loader.get_template('NearBeach/search/search.html')

    # Translate the include closed, from Python Boolean to JavaScript boolean
    if form.cleaned_data['include_closed']: #If exists and true
        include_closed = 'true'
    else:
        include_closed = 'false'

    # Context
    c = {
        'search_input': form.cleaned_data['search'],
        'include_closed': include_closed,
        'search_results': get_object_search_data(form),
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_data(request):
    """

    :param request:
    :return:
    """
    # Get the form data
    form = SearchObjectsForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Return the JSON data
    return JsonResponse(get_object_search_data(form))


@login_required(login_url='login',redirect_field_name="")
def search_customer(request):
    """

    :param request:
    :return:
    """
    t = loader.get_template('NearBeach/search/search_customers.html')

    # Get the first 50 customers
    customer_results = customer.objects.filter(
        is_deleted=False,
    ).order_by('customer_last_name','customer_first_name')[:50]

    c = {
        'customer_results': serializers.serialize('json',customer_results)
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_customer_data(request):
    # Get the data from request
    search_form = SearchForm(request.POST)

    # If there are errors - send 500
    if not search_form.is_valid():
        return HttpResponseBadRequest("There is an issue with the search functionality")

    # Get the base results
    customer_results = customer.objects.filter(is_deleted=False)

    # Split the space results - then apply the filter of each split value
    for split_row in search_form.cleaned_data['search'].split(' '):
        # Update the organisation results SQL
        customer_results = customer_results.filter(
            Q(customer_first_name__icontains=split_row) |
            Q(customer_last_name__icontains=split_row) |
            Q(organisation__organisation_name__icontains=split_row) # Might not work for freelancers
        )

    # Only have 50 results and order by alphabetical order
    customer_results.order_by('customer_last_name','customer_first_name')[:50]

    # Send back json data
    json_results = serializers.serialize('json', customer_results)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def search_group(request):
    """

    :param request:
    :return:
    """

    # ADD IN PERMISSIONS

    # Get template
    t = loader.get_template('NearBeach/search/search_groups.html')

    # Get user data
    group_results = group.objects.filter(
        is_deleted=False,
    ).order_by('group_name')[:25]

    # Get context
    c = {
        'group_results': serializers.serialize('json', group_results),
    }

    return HttpResponse(t.render(c, request))


# @require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_group_data(request):
    """

    :param request:
    :return:
    """

    # Obtain form data
    search_form = SearchForm(request.POST)
    if not search_form.is_valid():
        return HttpResponseBadRequest(search_form.errors)

    # Get the base group results
    group_results = group.objects.filter(
        is_deleted=False,
    )

    # Loop through the search results
    for split_row in search_form.cleaned_data['search'].split(' '):
        group_results = group_results.filter(
            group_name__icontains=split_row,
        )

    group_results.order_by('group_name')[:25]

    # Send back json data
    json_results = serializers.serialize('json', group_results)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def search_organisation(request):
    """

    :param request:
    :return:
    """
    t = loader.get_template('NearBeach/search/search_organisations.html')

    # Get the first 25 organisations
    organisation_results = organisation.objects.filter(
        is_deleted=False,
    ).order_by('organisation_name')[:25]

    c = {
        'organisation_results': serializers.serialize('json',organisation_results),
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_organisation_data(request):
    # Get the data from request
    search_form = SearchForm(request.POST)

    # If there are errors - send 500
    if not search_form.is_valid():
        return HttpResponseBadRequest("There is an issue with the search functionality")

    # Get the base results
    organisation_results = organisation.objects.filter(is_deleted=False)

    # Split the space results - then apply the filter of each split value
    for split_row in search_form.cleaned_data['search'].split(' '):
        # Update the organisation results SQL
        organisation_results = organisation_results.filter(
            organisation_name__icontains=split_row
        )

    # Only have 25 results and order by alphabetical order
    organisation_results = organisation_results.order_by('organisation_name')[:25]

    # Send back json data
    json_results = serializers.serialize('json', organisation_results)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def search_permission_set(request):
    """

    :param request:
    :return:
    """

    # Add permissions

    # Get template
    t = loader.get_template('NearBeach/search/search_permission_sets.html')

    # Get data
    permission_set_results = permission_set.objects.filter(
        is_deleted=False,
    ).order_by('permission_set_name')[:25]

    # Get context
    c = {
        'permission_set_results': serializers.serialize('json', permission_set_results),
    }

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_permission_set_data(request):
    """

    :param request:
    :return:
    """

    # Check user permission

    # Get form data
    search_form = SearchForm(request.POST)
    if not search_form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get base data
    permission_set_results = permission_set.objects.filter(
        is_deleted=False,
    )

    # Loop through the search results
    for split_row in search_form.cleaned_data['search'].split(' '):
        permission_set_results = permission_set_results.filter(
            permission_set_name__icontains=split_row,
        )

    permission_set_results.order_by('permission_set_name')[:25]

    # Send back json data
    json_results = serializers.serialize('json', permission_set_results)

    return HttpResponse(json_results, content_type='application/json')


@login_required(login_url='login',redirect_field_name="")
def search_user(request):
    """

    :param request:
    :return:
    """

    # Add permissions

    # Get template
    t = loader.get_template('NearBeach/search/search_users.html')

    # Get Data
    user_results = User.objects.filter(
    ).order_by('last_name','first_name')[:50]

    # Context
    c = {
        'user_results': serializers.serialize('json', user_results),
    }

    return HttpResponse(t.render(c, request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def search_user_data(request):
    # Get the data from request
    search_form = SearchForm(request.POST)

    # If there are errors - send 500
    if not search_form.is_valid():
        return HttpResponseBadRequest("There is an issue with the search functionality")

    # Get the base results
    user_results = User.objects.filter()

    # Split the space results - then apply the filter of each split value
    for split_row in search_form.cleaned_data['search'].split(' '):
        # Update the organisation results SQL
        user_results = user_results.filter(
            Q(first_name__icontains=split_row) |
            Q(last_name__icontains=split_row) |
            Q(username__icontains=split_row) |
            Q(email__icontains=split_row)
        )

    # Only have 50 results and order by alphabetical order
    user_results.order_by('last_name','first_name')[:50]

    # Send back json data
    json_results = serializers.serialize('json', user_results)

    return HttpResponse(json_results, content_type='application/json')


# Generated by Django 3.1.2 on 2021-03-10 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NearBeach', '0006_auto_20210227_1722'),
    ]

    operations = [
        migrations.AddField(
            model_name='kanban_card',
            name='kanban_card_description',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min
from NearBeach.forms import *
from NearBeach.views.tools.internal_functions import *
from NearBeach.user_permissions import return_user_permission_level
from django.db.models import Max

import json, urllib3


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def add_kanban_link(request,kanban_board_id,object_lookup):
    """

    :param request:
    :param destination:
    :param location_id:
    :return:
    """

    # CHECK USER PERMISSION LATER

    # Get form data and check
    form = AddKanbanLinkForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the newest card number id
    kanban_card_sort_number = get_max_sort_id(kanban_board_id, form)

    # Submit the kanban board link
    kanban_card_submit = kanban_card(
        change_user=request.user,
        kanban_board_id=kanban_board_id,
        kanban_column=form.cleaned_data['kanban_column'],
        kanban_level=form.cleaned_data['kanban_level'],
        kanban_card_sort_number=kanban_card_sort_number + 1,
    )

    # Check the data
    if object_lookup == 'project':
        kanban_card_submit.project = form.cleaned_data[object_lookup]
    elif object_lookup == 'task':
        kanban_card_submit.task = form.cleaned_data[object_lookup]
    elif object_lookup == 'requirement':
        kanban_card_submit.requirement = form.cleaned_data[object_lookup]

    kanban_card_submit.kanban_card_text = form.cleaned_data[object_lookup]

    # Save the data
    kanban_card_submit.save()

    # Send back the data we just created
    kanban_card_results = kanban_card.objects.get(kanban_card_id=kanban_card_submit.kanban_card_id)

    return HttpResponse(serializers.serialize('json',[kanban_card_results]),content_type='application/json')


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def check_kanban_board_name(request):
    """
    The following will get a list of all kanban boards with the same name. The idea is that each kanban board should
    have a unique name.
    :param request:
    :return:
    """

    # Get the form data
    form = CheckKanbanBoardName(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    kanban_board_results = kanban_board.objects.filter(
        is_deleted=False,
        kanban_board_name=form.cleaned_data['kanban_board_name'],
    )

    # Send back data
    return HttpResponse(serializers.serialize('json', kanban_board_results), content_type='application/json')


# Internal function
def get_max_sort_id(kanban_board_id,form):
    # Get the newest card number id
    kanban_card_sort_number = kanban_card.objects.filter(
        is_deleted=False,
        kanban_board_id=kanban_board_id,
        kanban_column=form.cleaned_data['kanban_column'],
        kanban_level=form.cleaned_data['kanban_level'],
    ).aggregate(Max('kanban_card_sort_number'))

    # If the card is new in that particular column & row - then we need to implement a sort number of 0
    if kanban_card_sort_number['kanban_card_sort_number__max'] == None:
        kanban_card_sort_number['kanban_card_sort_number__max'] = 0

    return kanban_card_sort_number['kanban_card_sort_number__max']


@login_required(login_url='login', redirect_field_name="")
def kanban_information(request, kanban_board_id):
    """

    :param request:
    :param kanban_board_id:
    :return:
    """

    # Check user's permissions

    # Get the kanban data
    kanban_board_results = kanban_board.objects.get(kanban_board_id=kanban_board_id)

    kanban_card_results = kanban_card.objects.filter(
        is_deleted=False,
        kanban_board_id=kanban_board_id,
    ).order_by('kanban_card_sort_number')

    column_results = kanban_column.objects.filter(
        is_deleted=False,
        kanban_board_id=kanban_board_id,
    )

    level_results = kanban_level.objects.filter(
        is_deleted=False,
        kanban_board_id=kanban_board_id,
    )

    # Get the template
    t = loader.get_template('NearBeach/kanban/kanban_information.html')

    # Context
    c = {
        'kanban_board_results': serializers.serialize('json', [kanban_board_results]),
        'kanban_card_results': serializers.serialize('json', kanban_card_results),
        'column_results': serializers.serialize('json', column_results),
        'kanban_board_id': kanban_board_id,
        'level_results': serializers.serialize('json', level_results),
    }

    return HttpResponse(t.render(c, request))


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def kanban_link_list(request, kanban_board_id, object_lookup):
    """

    :param request:
    :param kanban_board_id:
    :return:
    """

    # CHECK USER PERMISSIONS

    # Get a list of all existing cards
    existing_objects = kanban_card.objects.filter(
        is_deleted=False,
        kanban_board_id=kanban_board_id,
    )

    # Get the results we require
    if object_lookup == "Project":
        object_results = project.objects.filter(
            is_deleted=False,
        ).exclude(project_id__in=existing_objects.exclude(
            project_id__isnull=True
        ).values('project_id'))
    elif object_lookup == "Requirement":
        object_results = requirement.objects.filter(
            is_deleted=False,
        ).exclude(requirement_id__in=existing_objects.exclude(
            requirement_id__isnull=True
        ).values('requirement_id'))
    elif object_lookup == "Task":
        object_results = task.objects.filter(
            is_deleted=False,
        ).exclude(task_id__in=existing_objects.exclude(
            task_id__isnull=True,
        ).values('task_id'))
    else:
        return HttpResponseBadRequest("Sorry - there was an issue with your object lookup")

    return HttpResponse(serializers.serialize('json', object_results), content_type='application/json')


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def move_kanban_card(request, kanban_card_id):
    """

    :param request:
    :param kanban_board_id:
    :return:
    """

    # CHECK USER PERMISSIONS

    # Get the kanban card instance
    kanban_card_update = kanban_card.objects.get(kanban_card_id=kanban_card_id)

    # Get the form data
    form = MoveKanbanCardForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Update the card data
    kanban_card_update.kanban_column = form.cleaned_data['new_card_column']
    kanban_card_update.kanban_level = form.cleaned_data['new_card_level']
    kanban_card_update.kanban_card_sort_number = form.cleaned_data['new_card_sort_number']
    kanban_card_update.save()

    """
    Update the sort order
    ~~~~~~~~~~~~~~~~~~~~~
    
    If both the old and new level/column destination are the same, we take the difference between the two values
    Otherwise we apply two sort orders to both the old and the new
    """
    if form.cleaned_data['new_card_column'] == form.cleaned_data['old_card_column'] and \
            form.cleaned_data['new_card_level'] == form.cleaned_data['old_card_level']:
        # The card has stayed in the same column/level
        resort_array = kanban_card.objects.filter(
            Q(
                Q(
                    is_deleted=False,
                    kanban_card_sort_number__range=[
                        form.cleaned_data['old_card_sort_number'],
                        form.cleaned_data['new_card_sort_number']
                    ],
                ) |
                Q(
                    is_deleted=False,
                    kanban_card_sort_number__range=[
                        form.cleaned_data['new_card_sort_number'],
                        form.cleaned_data['old_card_sort_number']
                    ],
                )
            ) &
            ~Q(
                kanban_card_id=kanban_card_id,
            )
        ).order_by('kanban_card_sort_number')

        # Determine if we are using a positive or negative delta - using math
        delta = (-1) * (form.cleaned_data['new_card_sort_number'] > form.cleaned_data['old_card_sort_number']) + \
                (form.cleaned_data['new_card_sort_number'] < form.cleaned_data['old_card_sort_number'])

        # Send the data away to get manipulated
        update_sort_number(resort_array, delta)
    else:
        """
        The cards have been moved outside the origianl column/level. We need to update both the old and new location's
        sort orders.
        
        The old location will have a delta -1, to move the higher cards into the place left by the card
        
        The new location will have a delta +1, to move the higher cards away, to create a space for the card
        """
        old_resort_array = kanban_card.objects.filter(
            Q(
                is_deleted=False,
                kanban_card_sort_number__gte=form.cleaned_data['old_card_sort_number'],
            ) &
            ~Q(
                kanban_card_id=kanban_card_id,
            )
        ).order_by('kanban_card_sort_number')

        new_resort_array = kanban_card.objects.filter(
            Q(
                is_deleted=False,
                kanban_card_sort_number__gte=form.cleaned_data['new_card_sort_number'],
            ) &
            ~Q(
                kanban_card_id=kanban_card_id,
            )
        ).order_by('kanban_card_sort_number')

        update_sort_number(old_resort_array, -1)
        update_sort_number(new_resort_array, 1)

    return HttpResponse("")


@login_required(login_url='login', redirect_field_name="")
def new_kanban(request):
    """

    :param request:
    :return:
    """

    # Check user permissions

    # Get data
    group_results = group.objects.filter(
        is_deleted=False,
    )

    # Get tempalte
    t = loader.get_template('NearBeach/kanban/new_kanban.html')

    # Context
    c = {
        'group_results': serializers.serialize('json', group_results),
    }

    return HttpResponse(t.render(c, request))


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def new_kanban_card(request, kanban_board_id):
    """
    """

    # CHECK USER PERMISSIONS

    # Get the kanban instance
    kanban_instance = kanban_board.objects.get(kanban_board_id=kanban_board_id)

    # Get the form data
    form = NewKanbanCardForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the newest card number id
    kanban_card_sort_number = get_max_sort_id(kanban_board_id,form)

    # Save the data
    submit_kanban_card = kanban_card(
        change_user=request.user,
        kanban_board=kanban_instance,
        kanban_card_text=form.cleaned_data['kanban_card_text'],
        kanban_card_description=form.cleaned_data['kanban_card_description'],
        kanban_column=form.cleaned_data['kanban_column'],
        kanban_level=form.cleaned_data['kanban_level'],
        kanban_card_sort_number=kanban_card_sort_number + 1,
    )
    submit_kanban_card.save()

    # Send back the kanban card data
    kanban_card_results = kanban_card.objects.get(kanban_card_id=submit_kanban_card.kanban_card_id)
    return HttpResponse(serializers.serialize('json', [kanban_card_results]), content_type='application/json')


@require_http_methods(['POST'])
@login_required(login_url='login', redirect_field_name="")
def new_kanban_save(request):
    """

    :param request:
    :return:
    """

    # CHECK USERS PERMISSIONS

    # Check the user form
    form = NewKanbanForm(request.POST)
    if not form.is_valid():
        print(form.errors)
        return HttpResponseBadRequest(form.errors)

    # Create the kanban board
    submit_kanban_board = kanban_board(
        change_user=request.user,
        creation_user=request.user,
        kanban_board_name=form.cleaned_data['kanban_board_name'],
    )
    submit_kanban_board.save()

    # Get both lists
    column_title_list = request.POST.getlist("column_title")
    level_title_list = request.POST.getlist("level_title")

    # Loop through the column title list to save the titles
    for index, column_title in enumerate(column_title_list):
        submit_column_title = kanban_column(
            change_user=request.user,
            kanban_column_name=column_title,
            kanban_board=submit_kanban_board,
            kanban_column_sort_number=index,
        )
        submit_column_title.save()

    # Loop throuhg the level title list to save the titles
    for index, level_title in enumerate(level_title_list):
        submit_level_title = kanban_level(
            change_user=request.user,
            kanban_level_name=level_title,
            kanban_board=submit_kanban_board,
            kanban_level_sort_number=index,
        )
        submit_level_title.save()

    # Get the group list and apply the permissions
    group_list = request.POST.getlist("group_list")

    for single_group in group_list:
        # Get the group instance
        group_instance = group.objects.get(group_id=single_group)

        # Save the group instance against object assignment
        submit_object_assignment = object_assignment(
            group_id=group_instance,
            kanban_board=submit_kanban_board,
            change_user=request.user,
        )

        # Save
        submit_object_assignment.save()

    # Send back project_information URL
    return HttpResponse(reverse('kanban_information', args={submit_kanban_board.kanban_board_id}))


@login_required(login_url='login', redirect_field_name="")
@require_http_methods(['POST'])
def update_card(request):
    """
    The following function will update the card information sent through the form in POST
    """
    
    # ADD IN CHECKING USER PERMISSIONS

    # Get data and validate in the form
    form = KanbanCardForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    kanban_card_update = form.cleaned_data['kanban_card_id']
    kanban_card_update.kanban_card_text = form.cleaned_data['kanban_card_text']
    kanban_card_update.kanban_card_description = form.cleaned_data['kanban_card_description']
    kanban_card_update.save()

    return HttpResponse("")


# Internal Function
def update_sort_number(resort_array, delta):
    """
    The following function will loop through the resort array and apply the delta (which should be either a 1 or -1)
    :param resort_array:
    :param delta:
    :return:
    """
    for row in resort_array:
        row.kanban_card_sort_number = row.kanban_card_sort_number + delta
        row.save()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="NearBeach",
    version="0.27.8",
    author="Luke Christopher Clarke",
    author_email="luke@nearbeach.org",
    description="NearBeach - an open source project management tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/robotichead/NearBeach",
    packages=setuptools.find_packages(),
    install_requires=[
	'django',
	'simplejson',
	'pillow',
	'urllib3',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



from django.contrib.auth.decorators import login_required
from NearBeach.models import *
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.template import loader
from django.db.models import Sum, Q, Min
from NearBeach.forms import *
from NearBeach.user_permissions import return_user_permission_level
from django.views.decorators.http import require_http_methods


import json


@login_required(login_url='login',redirect_field_name='')
def customer_information(request,customer_id):
    """

    :param request:
    :param customer_id:
    :return:
    """
    # Find out if the user is read only - if they are send them to the read only

    # Get customer data
    customer_results = customer.objects.get(customer_id=customer_id)

    print("Customer Organisation id: %s " % customer_results.organisation_id)

    organisation_results = organisation.objects.filter(
        organisation_id=customer_results.organisation_id,
    )

    title_list = list_of_title.objects.filter(
        #is_deleted=False, # NEED TO RECONSTRUCT DATABASE TO GET THIS TO WORK!
    )

    # Get tempalte
    t = loader.get_template('NearBeach/customers/customer_information.html')

    # Context
    c = {
        'customer_results': serializers.serialize('json',[customer_results]),
        'organisation_results': serializers.serialize('json', organisation_results),
        'title_list': serializers.serialize('json',title_list),
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name='')
def customer_information_save(request,customer_id):
    """

    :param request:
    :param customer_id:
    :return:
    """
    # ADD IN USER PERMISSION CHECKS

    # Get form data
    form = CustomerForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Get the customer instance
    customer_results = customer.objects.get(customer_id=customer_id)

    # Update the fields
    customer_results.customer_title = form.cleaned_data['customer_title']
    customer_results.customer_first_name = form.cleaned_data['customer_first_name']
    customer_results.customer_last_name = form.cleaned_data['customer_last_name']
    customer_results.customer_email = form.cleaned_data['customer_email']
    # customer_results.organisation = form.cleaned_data['organisation'] # Does not need updating!

    # Save
    customer_results.save()

    # Return back a sucessful statement
    return HttpResponse("Success")


@login_required(login_url='login',redirect_field_name="")
def new_customer(request):
    """

    :param request:
    :return:
    """
    # Get user permission

    # Get data
    title_list = list_of_title.objects.filter(
        is_deleted=False,
    )

    # Get templates
    t = loader.get_template('NearBeach/customers/new_customers.html')

    # Get Context
    c = {
        'title_list': serializers.serialize('json',title_list),
    }

    return HttpResponse(t.render(c,request))


@require_http_methods(['POST'])
@login_required(login_url='login',redirect_field_name="")
def new_customer_save(request):
    """

    :param reqeust:
    :return:
    """

    # CHECK USER PERMISSIONS -- NEED TO ADD THIS IN!

    # Get Form data
    form = NewCustomerForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(form.errors)

    # Save the data
    customer_submit = customer(
        change_user=request.user,
        customer_title=form.cleaned_data['customer_title'],
        customer_first_name=form.cleaned_data['customer_first_name'],
        customer_last_name=form.cleaned_data['customer_last_name'],
        customer_email=form.cleaned_data['customer_email'],
    )

    # If the organisation is not null in the form - set the data
    if form.cleaned_data['organisation']:
        customer_submit.organisation = form.cleaned_data['organisation']

    # Save the data
    customer_submit.save()

    # Send back the URL to the new customer
    return HttpResponse(reverse('customer_information', args={customer_submit.customer_id}))